"""
    Source: https://github.com/hallazzang/asyncio-throttle

    Mods:
        - add loop arg
        - add size_limit to support throttling by size
"""

import time
import math
import asyncio
import logging
from collections import deque

log = logging.getLogger(__name__)


class Throttler:
    def __init__(
        self,
        rate_limit=None,
        size_limit=None,
        period=1.0,
        retry_interval=0.05,
        loop=None,
    ):
        self.rate_limit = rate_limit
        self.size_limit = size_limit
        self.period = period
        self.loop = loop
        self.retry_interval = retry_interval

        self._task_logs = deque()

        self.size = None

    def flush(self):
        now = time.time()
        while self._task_logs:
            if now - self._task_logs[0][0] > self.period:
                self._task_logs.popleft()
            else:
                break

    def is_below_rate(self):

        if self.rate_limit:
            below_rate_requests = len(self._task_logs) < self.rate_limit

            if not below_rate_requests:
                return False

        if self.size_limit is None or not self._task_logs:
            return True

        size = sum([x[1] for x in self._task_logs])

        period = time.time() - self._task_logs[0][0]

        period_used_ratio = (self.period - period) / self.period

        remaining = self.size_limit - math.ceil(size * period_used_ratio)

        # log.debug("rate check: size={} requested={} period={} period_used_ratio={} remaining={}".format(size, self.size, round(period,3), round(period_used_ratio, 2), round(remaining,2)))

        return self.size <= remaining

    async def acquire(self):

        while True:
            self.flush()
            if self.is_below_rate():
                break
            await asyncio.sleep(
                self.retry_interval,
                loop=self.loop if self.loop else asyncio.get_event_loop(),
            )

        self._task_logs.append((time.time(), self.size))

    def __call__(self, size=1):
        self.size = size
        return self

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        pass



import asyncio
import aiobotocore
import logging
from async_timeout import timeout
from asyncio import CancelledError
from botocore.exceptions import ClientError
from botocore.config import Config

from . import exceptions


log = logging.getLogger(__name__)


class Base:
    def __init__(self, stream_name, loop=None, endpoint_url=None, region_name=None):

        self.stream_name = stream_name
        self.loop = loop if loop else asyncio.get_event_loop()

        self.endpoint_url = endpoint_url
        self.region_name = region_name

        self.client = None
        self.shards = None

        self.stream_status = None

    async def __aenter__(self):

        log.info(
            "creating client with {}".format(
                self.endpoint_url if self.endpoint_url else "AWS default endpoint"
            )
        )

        session = aiobotocore.get_session(loop=self.loop)

        # Note: max_attempts = 0
        # Boto RetryHandler only handles these errors:
        #  GENERAL_CONNECTION_ERROR => ConnectionError, ConnectionClosedError, ReadTimeoutError, EndpointConnectionError
        # Still have to handle ClientError anyway~

        self.client = session.create_client(
            "kinesis",
            endpoint_url=self.endpoint_url,
            region_name=self.region_name,
            config=Config(
                connect_timeout=5, read_timeout=90, retries={"max_attempts": 0}
            ),
        )

        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def get_stream_description(self):

        try:
            return (await self.client.describe_stream(StreamName=self.stream_name))[
                "StreamDescription"
            ]
        except ClientError as err:
            code = err.response["Error"]["Code"]
            if code == "ResourceNotFoundException":
                raise exceptions.StreamDoesNotExist(
                    "Stream '{}' does not exist".format(self.stream_name)
                ) from None

    async def start(self):

        log.debug("Checking stream '{}' is active".format(self.stream_name))

        async with timeout(60, loop=self.loop) as cm:
            try:
                while True:
                    stream_info = await self.get_stream_description()
                    stream_status = stream_info["StreamStatus"]

                    if stream_status == "ACTIVE":
                        self.stream_status = stream_status
                        break

                    if stream_status in ["CREATING", "UPDATING"]:
                        await asyncio.sleep(0.25)

                    else:
                        raise exceptions.StreamStatusInvalid(
                            "Stream '{}' is {}".format(self.stream_name, stream_status)
                        )
            except CancelledError:
                pass

            else:
                self.shards = stream_info["Shards"]

        if cm.expired:
            raise exceptions.StreamStatusInvalid(
                "Stream '{}' is still {}".format(self.stream_name, stream_status)
            )



import logging
import math
from collections import namedtuple
from .exceptions import ValidationError
from .exceptions import ExceededPutLimit

log = logging.getLogger(__name__)

OutputItem = namedtuple("OutputItem", ["size", "n", "data"])


class BaseAggregator:
    def __init__(self, max_size=None):

        if not max_size:
            max_size = 1024

        put_units = math.floor(max_size / 25)

        if put_units <= 0:
            raise ValidationError(
                "max_size is too low. Should be at least one PUT Payload Unit (25Kb)"
            )

        if put_units > 40:
            raise ValidationError(
                "max_size is too high. Should be no higher than 40x PUT Payload Units (25Kb)"
            )

        self.max_bytes = put_units * 25 * 1024

        log.debug(
            "setting max_bytes to {} ({} PUT Payload Units (25kb))".format(
                self.max_bytes, put_units
            )
        )

        self.buffer = []
        self.size = 0

    def validate_size(self, size):
        if size > self.max_bytes:
            raise ExceededPutLimit("Put of {} bytes exceeded 1MB limit".format(size))

    def parse(self, data):
        yield self.deserialize(data)


class SimpleAggregator(BaseAggregator):
    """
    Simple Aggregator (Does NOT aggregate)
    Sends a single record only (high inefficient)
    """

    def has_items(self):
        return False

    def add_item(self, item):
        output = self.serialize(item)
        size = len(output)

        self.validate_size(size)

        yield OutputItem(size=size, n=1, data=output)


class Aggregator(BaseAggregator):
    """
    Aggregator
    Sends an aggregated record
    """

    def has_items(self):
        return self.size > 0

    def get_header_size(self, data):
        raise NotImplementedError()

    def add_item(self, item):
        output = self.serialize(item)
        size = len(output)

        self.validate_size(size)

        header_size = self.get_header_size(output)

        if size + self.size + header_size < self.max_bytes:
            self.buffer.append((size, output))
            self.size += size + header_size

        else:
            log.debug(
                "Adding (full) item to queue with {} individual records with size of {} kb".format(
                    len(self.buffer), round(self.size / 1024)
                )
            )
            yield OutputItem(size=self.size, n=len(self.buffer), data=self.output())
            self.buffer = [(size, output)]
            self.size = 1

    def get_items(self):
        log.debug(
            "Adding (partial) item to queue with {} individual records with size of {} kb".format(
                len(self.buffer), len(self.buffer), round(self.size / 1024)
            )
        )
        yield OutputItem(size=self.size, n=len(self.buffer), data=self.output())
        self.buffer = []
        self.size = 0


class NewlineAggregator(Aggregator):
    def get_header_size(self, output):
        return 1

    def output(self):
        return b"\n".join([x[1] for x in self.buffer] + [b""])

    def parse(self, data):
        for row in data.split(b"\n"):
            if row:
                yield self.deserialize(row)


class NetstringAggregator(Aggregator):
    """
    Netstring Aggregation
    Framing = {x} bytes (ascii int for size) + 1 byte (":") + data + trailing ","
    See: https://en.wikipedia.org/wiki/Netstring
    """

    def get_header_size(self, output):
        return len(str(len(output))) + 2

    def output(self):
        frame = []

        for size, data in self.buffer:
            frame.append(str(size).encode("ascii"))
            frame.append(b":")
            frame.append(data)
            frame.append(b",")

        return b"".join(frame)

    def parse(self, data):

        i = 0
        length = len(data)

        while True:
            header_offset = data[i:].index(b":")
            size = int(data[i : i + header_offset].decode("ascii"))
            item = data[i + header_offset + 1 : i + header_offset + 1 + size]
            yield self.deserialize(item)

            i += header_offset + size + 2
            if i == length:
                break



from .producer import Producer
from .consumer import Consumer
from .processors import (
    StringProcessor,
    JsonProcessor,
    JsonLineProcessor,
    MsgpackProcessor,
)
from .serializers import StringSerializer, JsonSerializer, MsgpackSerializer
from .checkpointers import MemoryCheckPointer, RedisCheckPointer
from .aggregators import SimpleAggregator, NewlineAggregator, NetstringAggregator



import asyncio
import logging
from datetime import datetime, timezone
from aiohttp import ClientConnectionError
from asyncio import TimeoutError
from asyncio.queues import QueueEmpty
from botocore.exceptions import ClientError
from .utils import Throttler
from .base import Base
from .checkpointers import MemoryCheckPointer
from .processors import JsonProcessor

log = logging.getLogger(__name__)


class ShardStats:
    def __init__(self):
        self._throttled = 0
        self._success = 0

    def succeded(self):
        self._success += 1

    def throttled(self):
        self._throttled += 1

    def to_data(self):
        return {"throttled": self._throttled, "success": self._success}


class Consumer(Base):
    def __init__(
        self,
        stream_name,
        loop=None,
        endpoint_url=None,
        region_name=None,
        max_queue_size=10000,
        max_shard_consumers=None,
        record_limit=10000,
        sleep_time_no_records=2,
        iterator_type="TRIM_HORIZON",
        shard_fetch_rate=1,
        checkpointer=None,
        processor=None,
    ):

        super(Consumer, self).__init__(
            stream_name, loop=loop, endpoint_url=endpoint_url, region_name=region_name
        )

        self.queue = asyncio.Queue(maxsize=max_queue_size, loop=self.loop)

        self.sleep_time_no_records = sleep_time_no_records

        self.max_shard_consumers = max_shard_consumers

        self.record_limit = record_limit

        self.is_fetching = True

        self.checkpointer = checkpointer if checkpointer else MemoryCheckPointer()

        self.processor = processor if processor else JsonProcessor()

        self.iterator_type = iterator_type

        self.fetch_task = None

        self.shard_fetch_rate = shard_fetch_rate

    def __aiter__(self):
        return self

    async def close(self):
        log.debug("Closing..")

        await self.flush()

        if self.fetch_task:
            self.fetch_task.cancel()
            self.fetch_task = None

        if self.checkpointer:
            await self.checkpointer.close()

        await self.client.close()

    async def flush(self):

        self.is_fetching = False

        if not self.shards:
            return

        # Wait for shard fetches to finish
        # todo: use gather
        for shard in self.shards:
            if shard.get("fetch"):
                if not shard["fetch"].done():
                    await shard["fetch"]

    async def _fetch(self):
        while self.is_fetching:
            # Ensure fetch is performed at most 5 times per second (the limit per shard)
            await asyncio.sleep(0.2, loop=self.loop)
            try:
                await self.fetch()
            except asyncio.CancelledError:
                pass
            except Exception as e:
                log.exception(e)

    async def fetch(self):

        if not self.is_fetching:
            return

        # todo: check for/handle new shards

        shards_in_use = [
            s for s in self.shards if self.checkpointer.is_allocated(s["ShardId"])
        ]

        # log.debug("shards in use: {}".format([s["ShardId"] for s in shards_in_use]))

        for shard in self.shards:

            if not self.is_fetching:
                break

            if not self.checkpointer.is_allocated(shard["ShardId"]):
                if (
                    self.max_shard_consumers
                    and len(shards_in_use) >= self.max_shard_consumers
                ):
                    continue

                if self.checkpointer is None:
                    log.debug("Marking shard in use {}".format(shard["ShardId"]))
                    shard["ShardIterator"] = await self.get_shard_iterator(
                        shard_id=shard["ShardId"]
                    )

                else:
                    success, checkpoint = await self.checkpointer.allocate(
                        shard["ShardId"]
                    )

                    if not success:
                        log.debug(
                            "Shard in use. Could not assign shard {} to checkpointer[{}]".format(
                                shard["ShardId"], self.checkpointer.get_ref()
                            )
                        )
                        continue

                    log.debug(
                        "Marking shard in use {} by checkpointer[{}] @ {}".format(
                            shard["ShardId"], self.checkpointer.get_ref(), checkpoint
                        )
                    )

                    shard["ShardIterator"] = await self.get_shard_iterator(
                        shard_id=shard["ShardId"], last_sequence_number=checkpoint
                    )

                if "ShardIterator" in shard:
                    shard["stats"] = ShardStats()
                    shard["throttler"] = Throttler(
                        rate_limit=self.shard_fetch_rate, period=1, loop=self.loop
                    )
                    shards_in_use.append(shard)

                    log.debug("Shard count now at {}".format(len(shards_in_use)))

            if shard.get("fetch"):
                if shard["fetch"].done():
                    result = shard["fetch"].result()

                    if not result:
                        shard["fetch"] = None
                        continue

                    records = result["Records"]

                    if records:
                        log.debug(
                            "Shard {} got {} records".format(
                                shard["ShardId"], len(records)
                            )
                        )

                        total_items = 0
                        for row in result["Records"]:
                            for n, output in enumerate(
                                self.processor.parse(row["Data"])
                            ):
                                await self.queue.put(output)
                            total_items += n + 1

                        # Get approx minutes behind..
                        last_arrival = records[-1].get("ApproximateArrivalTimestamp")
                        if last_arrival:
                            last_arrival = round(
                                (
                                    (
                                        datetime.now(timezone.utc) - last_arrival
                                    ).total_seconds()
                                    / 60
                                )
                            )

                            log.debug(
                                "Shard {} added {} items from {} records. Consumer is {}m behind".format(
                                    shard["ShardId"],
                                    total_items,
                                    len(records),
                                    last_arrival,
                                ),
                                extra={"consumer_behind_m": last_arrival},
                            )

                        else:
                            # ApproximateArrivalTimestamp not available in kinesis-lite
                            log.debug(
                                "Shard {} added {} items from {} records".format(
                                    shard["ShardId"], total_items, len(records)
                                )
                            )

                        # Add checkpoint record
                        last_record = result["Records"][-1]
                        await self.queue.put(
                            {
                                "__CHECKPOINT__": {
                                    "ShardId": shard["ShardId"],
                                    "SequenceNumber": last_record["SequenceNumber"],
                                }
                            }
                        )

                        shard["LastSequenceNumber"] = last_record["SequenceNumber"]

                    else:
                        log.debug(
                            "Shard {} caught up, sleeping {}s".format(
                                shard["ShardId"], self.sleep_time_no_records
                            )
                        )
                        await asyncio.sleep(self.sleep_time_no_records, loop=self.loop)

                    if not result["NextShardIterator"]:
                        raise NotImplementedError("Shard is closed?")

                    shard["ShardIterator"] = result["NextShardIterator"]

                    shard["fetch"] = None

                else:
                    # log.debug("shard {} fetch in progress..".format(shard['ShardId']))
                    continue

            if "ShardIterator" in shard and shard["ShardIterator"] is not None:
                shard["fetch"] = self.loop.create_task(self.get_records(shard=shard))

    async def get_records(self, shard):

        # Note: "This operation has a limit of five transactions per second per account."

        async with shard["throttler"]:
            # log.debug("get_records shard={}".format(shard['ShardId']))

            try:

                result = await self.client.get_records(
                    ShardIterator=shard["ShardIterator"], Limit=self.record_limit
                )

                shard["stats"].succeded()
                return result

            except ClientConnectionError as e:
                log.warning("Connection error {}. sleeping..".format(e))
                await asyncio.sleep(3, loop=self.loop)
            except TimeoutError as e:
                log.warning("Timeout {}. sleeping..".format(e))
                await asyncio.sleep(3, loop=self.loop)

            except ClientError as e:
                code = e.response["Error"]["Code"]
                if code == "ProvisionedThroughputExceededException":
                    log.warning(
                        "{} hit ProvisionedThroughputExceededException".format(
                            shard["ShardId"]
                        )
                    )
                    shard["stats"].throttled()
                    # todo: control the throttle ?
                    await asyncio.sleep(0.25, loop=self.loop)

                elif code == "ExpiredIteratorException":
                    log.warning(
                        "{} hit ExpiredIteratorException".format(shard["ShardId"])
                    )

                    shard["ShardIterator"] = await self.get_shard_iterator(
                        shard_id=shard["ShardId"],
                        last_sequence_number=shard.get("LastSequenceNumber"),
                    )

                else:
                    log.warning("ClientError {}. sleeping..".format(code))
                    await asyncio.sleep(3, loop=self.loop)

            except Exception as e:
                log.warning("Unknown error {}. sleeping..".format(e))
                await asyncio.sleep(3, loop=self.loop)

        # Connection or other issue
        return None

    async def get_shard_iterator(self, shard_id, last_sequence_number=None):

        log.debug(
            "getting shard iterator for {} @ {}".format(
                shard_id,
                last_sequence_number if last_sequence_number else self.iterator_type,
            )
        )

        params = {
            "StreamName": self.stream_name,
            "ShardId": shard_id,
            "ShardIteratorType": "AFTER_SEQUENCE_NUMBER"
            if last_sequence_number
            else self.iterator_type,
        }

        if last_sequence_number:
            params["StartingSequenceNumber"] = last_sequence_number

        response = await self.client.get_shard_iterator(**params)
        return response["ShardIterator"]

    async def start_consumer(self, wait_iterations=10, wait_sleep=0.25):

        await self.start()

        # Start task to fetch periodically
        self.fetch_task = asyncio.Task(self._fetch(), loop=self.loop)

        # Wait a while until we have some results
        for i in range(0, wait_iterations):
            if self.fetch_task and self.queue.qsize() == 0:
                await asyncio.sleep(wait_sleep, loop=self.loop)

        log.debug("start_consumer completed.. queue size={}".format(self.queue.qsize()))

    async def __anext__(self):

        if not self.shards:
            await self.start_consumer()

        while True:
            try:
                item = self.queue.get_nowait()

                if item and "__CHECKPOINT__" in item:
                    if self.checkpointer:
                        await self.checkpointer.checkpoint(
                            item["__CHECKPOINT__"]["ShardId"],
                            item["__CHECKPOINT__"]["SequenceNumber"],
                        )
                        continue
                    else:
                        continue

                return item

            except QueueEmpty:
                log.debug("Queue empty..")
                await asyncio.sleep(self.sleep_time_no_records, loop=self.loop)
                raise StopAsyncIteration



import asyncio
import logging
import time
import math
from aiohttp import ClientConnectionError

from asyncio.queues import QueueEmpty

from .utils import Throttler
from botocore.exceptions import ClientError

from .base import Base
from . import exceptions
from .processors import JsonProcessor

log = logging.getLogger(__name__)


class Producer(Base):
    def __init__(
        self,
        stream_name,
        loop=None,
        endpoint_url=None,
        region_name=None,
        buffer_time=0.5,
        put_rate_limit_per_shard=1000,
        put_bandwidth_limit_per_shard=1024,
        after_flush_fun=None,
        batch_size=500,
        max_queue_size=10000,
        processor=None,
    ):

        super(Producer, self).__init__(
            stream_name, loop=loop, endpoint_url=endpoint_url, region_name=region_name
        )

        self.buffer_time = buffer_time

        self.processor = processor if processor else JsonProcessor()

        self.queue = asyncio.Queue(maxsize=max_queue_size, loop=self.loop)

        self.batch_size = batch_size

        # A single shard can ingest up to 1 MiB of data per second (including partition keys)
        # or 1,000 records per second for writes
        self.put_rate_limit_per_shard = put_rate_limit_per_shard
        self.put_rate_throttle = None
        self.put_bandwidth_limit_per_shard = put_bandwidth_limit_per_shard
        self.put_bandwidth_throttle = None

        if put_bandwidth_limit_per_shard > 1024:
            log.warning(
                (
                    "Put bandwidth {}kb exceeds 1024kb. Expect throughput errors..".format(
                        put_bandwidth_limit_per_shard
                    )
                )
            )

        self.flush_task = asyncio.Task(self._flush(), loop=self.loop)
        self.is_flushing = False
        self.after_flush_fun = after_flush_fun

        # Keep flush task looping while active
        self.active = True

        # keep track of these (used by unit test only)
        self.throughput_exceeded_count = 0

    async def create_stream(self, shards=1, ignore_exists=True):

        log.debug(
            "Creating (or ignoring) stream {} with {} shards".format(
                self.stream_name, shards
            )
        )

        if shards < 1:
            raise Exception("Min shard count is one")

        try:
            await self.client.create_stream(
                StreamName=self.stream_name, ShardCount=shards
            )
        except ClientError as err:
            code = err.response["Error"]["Code"]

            if code == "ResourceInUseException":
                if not ignore_exists:
                    raise exceptions.StreamExists(
                        "Stream '{}' exists, cannot create it".format(self.stream_name)
                    ) from None
            elif code == "LimitExceededException":
                raise exceptions.StreamShardLimit(
                    "Stream '{}' exceeded shard limit".format(self.stream_name)
                )
            else:
                raise

    def set_put_rate_throttle(self):
        self.put_rate_throttle = Throttler(
            rate_limit=self.put_rate_limit_per_shard * len(self.shards),
            period=1,
            loop=self.loop,
        )
        self.put_bandwidth_throttle = Throttler(
            # kb per second. Go below a bit to avoid hitting the threshold
            size_limit=self.put_bandwidth_limit_per_shard * len(self.shards),
            period=1,
            loop=self.loop,
        )

    async def put(self, data):

        if not self.stream_status == "ACTIVE":
            await self.start()
            self.set_put_rate_throttle()

        if self.queue.qsize() >= self.batch_size:
            await self.flush()

        for output in self.processor.add_item(data):
            await self.queue.put(output)

    async def close(self):
        self.active = False
        # Wait till task completes
        await self.flush_task

        # final flush (probably not required but no harm)
        await asyncio.shield(self.flush())

        await self.client.close()

    async def _flush(self):
        while self.active:
            await asyncio.sleep(self.buffer_time, loop=self.loop)
            if not self.is_flushing:
                await asyncio.shield(self.flush())

    async def flush(self):

        if self.is_flushing:
            log.debug("Flush already in progress, ignoring..")
            return

        self.is_flushing = True

        if self.processor.has_items():
            for output in self.processor.get_items():
                await self.queue.put(output)

        # Overflow in case we run into trouble with a batch
        overflow = []

        while True:

            total_records = 0
            max_size = 0
            total_size = 0

            if self.queue.qsize() > 0 or len(overflow) > 0:
                log.debug(
                    "flush queue={} overflow={}".format(
                        self.queue.qsize(), len(overflow)
                    )
                )

            items = []

            num = min(self.batch_size, self.queue.qsize() + len(overflow))

            for _ in range(num):
                async with self.put_rate_throttle:

                    if overflow:
                        item = overflow.pop()

                    else:
                        try:
                            item = self.queue.get_nowait()
                        except QueueEmpty:
                            break

                    size_kb = math.ceil(item[0] / 1024)

                    max_size += size_kb

                    if max_size > 1024:
                        overflow.append(item)
                        break

                    total_size += size_kb
                    total_records += item[1]

                    async with self.put_bandwidth_throttle(size=total_size):
                        pass

                    items.append(item)

            if not items:
                break

            log.debug(
                "doing flush with {} record ({} items) @ {} kb".format(
                    len(items), total_records, total_size
                )
            )

            try:

                # todo: custom partition key
                result = await self.client.put_records(
                    Records=[
                        {
                            "Data": item.data,
                            "PartitionKey": "{0}{1}".format(time.perf_counter(), time.time()),
                        }
                        for item in items
                    ],
                    StreamName=self.stream_name,
                )

            except ClientError as err:

                code = err.response["Error"]["Code"]

                if code == "ValidationException":
                    if (
                        "must have length less than or equal"
                        in err.response["Error"]["Message"]
                    ):
                        log.warning(
                            "Batch size {} exceeded the limit. retrying with less".format(
                                len(items)
                            )
                        )
                        overflow = items

                        existing_batch_size = self.batch_size
                        self.batch_size -= round(self.batch_size / 10)

                        # Must be small batch of big items, take at least one out..
                        if existing_batch_size == self.batch_size:
                            self.batch_size -= 1

                        continue
                    else:
                        raise
                elif code == "ResourceNotFoundException":
                    raise exceptions.StreamDoesNotExist(
                        "Stream '{}' does not exist".format(self.stream_name)
                    ) from None
                else:
                    raise
            except ClientConnectionError:
                log.warning("Connection error. sleeping..")
                overflow = items
                await asyncio.sleep(3, loop=self.loop)
                continue
            except Exception:
                raise
            else:

                if result["FailedRecordCount"]:

                    errors = list(
                        set(
                            [
                                r.get("ErrorCode")
                                for r in result["Records"]
                                if r.get("ErrorCode")
                            ]
                        )
                    )

                    if not errors:
                        raise exceptions.UnknownException(
                            "Failed to put records but no errorCodes return in results"
                        )

                    if "ProvisionedThroughputExceededException" in errors:
                        log.warning(
                            "Throughput exceeded ({} records failed, added back..), pausing for 0.25s..".format(
                                result["FailedRecordCount"]
                            )
                        )

                        self.throughput_exceeded_count += 1

                        for i, record in enumerate(result["Records"]):
                            if "ErrorCode" in record:
                                overflow.append(items[i])

                        # log.debug("items={} overflow={}".format(len(items), len(overflow)))

                        await asyncio.sleep(0.25, loop=self.loop)

                    else:
                        raise exceptions.UnknownException(
                            "Failed to put records due to: {}".format(", ".join(errors))
                        )

                else:

                    if self.after_flush_fun:
                        await self.after_flush_fun(items)

        self.is_flushing = False



from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="async-kinesis",
    description="AsyncIO Kinesis Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.3.0",
    url="https://github.com/hampsterx/async-kinesis",
    author="hampsterx",
    author_email="tim.vdh@gmail.com",
    license='Apache2',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: Apache Software License',
    ],
    packages=["kinesis"],
    install_requires=[
        "aiobotocore>=0.11.0",
        "async-timeout==3.0.1",
        "asyncio-throttle==0.1.1",
    ],
    extras_require={
        "redis": ["aredis>=1.1.7"],
        "msgpack": ["msgpack>=0.6.1"]
    }
)